import time


class Model():
    def __init__(self, queue):
        self.turn_timeout = 0.4
        self.turn_start_time = 0
        self.world = None
        self.queue = queue

    def handle_init_message(self, message):
        init_info = message[Constants.KEY_ARGS][0]
        map_data = message[Constants.KEY_ARGS][1]
        map1 = Map(init_info[Constants.INFO_KEY_MAPSIZE], map_data)
        self.world = World(self, init_info, map1)
        # TODO static diff

    def handle_turn_message(self, message):
        self.turn_start_time = time.time()
        self.world.turn = message[Constants.KEY_ARGS][0]
        turn_data = message[Constants.KEY_ARGS][1]
        for static_data in turn_data[Constants.KEY_STATICS]:
            self.world.set_static_change(static_data)
        self.world.clear_dynamics()
        for dynamic_data in turn_data[Constants.KEY_DYNAMICS]:
            if dynamic_data[Constants.GAME_OBJECT_KEY_TYPE] == Constants.GAME_OBJECT_TYPE_CELL:
                self.world.add_cell(dynamic_data)
            else:
                pass
                # TODO Nothing yet
        # TODO set transients

    def turn_remaining_time(self):
        passed = time.time() - self.turn_start_time
        return self.turn_timeout - passed


class World():
    def __init__(self, model, init_info, map1):
        self.model = model
        self.teams = init_info[Constants.INFO_KEY_TEAMS]
        self.my_name = init_info[Constants.INFO_KEY_YOUR_INFO][Constants.KEY_NAME]
        self.my_id = init_info[Constants.INFO_KEY_YOUR_INFO][Constants.GAME_OBJECT_KEY_ID]
        self.map_size = init_info[Constants.INFO_KEY_MAPSIZE]
        self.map = map1
        self.turn = init_info[Constants.GAME_OBJECT_KEY_TURN]
        self.all_cells = []
        self.my_cells = []
        self.enemy_cells = []

    def clear_dynamics(self):
        self.all_cells = []
        self.my_cells = []
        self.enemy_cells = []

    def add_cell(self, cell_data):
        cell = Cell(self.model, cell_data[Constants.GAME_OBJECT_KEY_ID], cell_data[Constants.GAME_OBJECT_KEY_POSITION],
                    cell_data[Constants.KEY_OTHER][Constants.GAME_OBJECT_KEY_TEAM_ID],
                    cell_data[Constants.KEY_OTHER][Constants.CELL_KEY_ENERGY])
        self.all_cells.append(cell)
        if cell_data[Constants.KEY_OTHER][Constants.GAME_OBJECT_KEY_TEAM_ID] == self.my_id:
            self.my_cells.append(cell)
        else:
            self.enemy_cells.append(cell)

    def set_static_change(self, static_data):
        if Block.is_block_type(static_data[Constants.GAME_OBJECT_KEY_TYPE]):
            self.map.set_change(static_data)
        else:
            # TODO NOTHING YET!
            pass


class Cell():
    def __init__(self, model, m_id, pos, team_id, energy):
        self.model = model
        self.id = m_id
        self.pos = pos
        self.teamId = team_id
        self.energy = energy

    def move(self, direction):
        event = Event(Event.TYPE_MOVE, self.id)
        event.add_arg(direction)
        self.model.queue.put(event.to_message())

    def gain_resource(self):
        event = Event(Event.TYPE_GAIN_RESOURCE, self.id)
        self.model.queue.put(event.to_message())

    def mitosis(self):
        event = Event(Event.TYPE_MITOSIS, self.id)
        self.model.queue.put(event.to_message())


class Map():
    def __init__(self, map_size, map_data):
        self.blocks = [[None for i in range(map_size["width"])] for j in range(map_size["height"])]
        for static_data in map_data:
            self.set_change(static_data)

    def set_change(self, data):
        if not Block.is_block_type(data[Constants.GAME_OBJECT_KEY_TYPE]):
            return False
        pos = data[Constants.GAME_OBJECT_KEY_POSITION]
        if not self.at(pos):
            self.blocks[pos["y"]][pos["x"]] = Block(data)
        else:
            self.at(pos).set_change(data)
        return True

    def at(self, pos):
        return self.blocks[pos["y"]][pos["x"]]

    def get_next_pos(self, direction, pos):
        x = pos["x"]
        y = pos["y"]
        if x % 2 == 1:
            if direction == Constants.Directions.NORTH:
                return {"x": x, "y": y + 1}
            elif direction == Constants.Directions.SOUTH:
                return {"x": x, "y": y - 1}
            elif direction == Constants.Directions.NORTH_EAST:
                return {"x": x + 1, "y": y}
            elif direction == Constants.Directions.NORTH_WEST:
                return {"x": x - 1, "y": y}
            elif direction == Constants.Directions.SOUTH_EAST:
                return {"x": x + 1, "y": y - 1}
            elif direction == Constants.Directions.SOUTH_WEST:
                return {"x": x - 1, "y": y - 1}
            else:
                return None
        else:
            if direction == Constants.Directions.NORTH:
                return {"x": x, "y": y + 1}
            elif direction == Constants.Directions.SOUTH:
                return {"x": x, "y": y - 1}
            elif direction == Constants.Directions.NORTH_EAST:
                return {"x": x + 1, "y": y + 1}
            elif direction == Constants.Directions.NORTH_WEST:
                return {"x": x - 1, "y": y + 1}
            elif direction == Constants.Directions.SOUTH_EAST:
                return {"x": x + 1, "y": y}
            elif direction == Constants.Directions.SOUTH_WEST:
                return {"x": x - 1, "y": y}
            else:
                return None


class Block():
    BLOCK_TYPES = [
        "none",
        "normal",
        "mitosis",
        "resource",
        "impassible",
    ]

    def __init__(self, data):
        self.pos = data[Constants.GAME_OBJECT_KEY_POSITION]
        self.height = data[Constants.KEY_OTHER][Constants.BLOCK_KEY_HEIGHT]
        self.resource = data[Constants.KEY_OTHER][Constants.BLOCK_KEY_RESOURCE]
        self.turn = data[Constants.GAME_OBJECT_KEY_TURN]
        self.type = data[Constants.GAME_OBJECT_KEY_TYPE]
        self.id = data[Constants.GAME_OBJECT_KEY_ID]

    def set_change(self, data):
        self.pos = data[Constants.GAME_OBJECT_KEY_POSITION]
        self.height = data[Constants.KEY_OTHER][Constants.BLOCK_KEY_HEIGHT]
        self.resource = data[Constants.KEY_OTHER][Constants.BLOCK_KEY_RESOURCE]
        self.turn = data[Constants.GAME_OBJECT_KEY_TURN]
        self.type = data[Constants.GAME_OBJECT_KEY_TYPE]
        self.id = data[Constants.GAME_OBJECT_KEY_ID]

    @classmethod
    def is_block_type(cls, type1):
        if type1 in Block.BLOCK_TYPES:
            return True
        else:
            return False


class Constants():
    class Directions:
        NORTH = "NORTH"
        NORTH_EAST = "NORTH_EAST"
        SOUTH_EAST = "SOUTH_EAST"
        SOUTH = "SOUTH"
        SOUTH_WEST = "SOUTH_WEST"
        NORTH_WEST = "NORTH_WEST"

    KEY_ARGS = "args"
    KEY_NAME = "name"
    KEY_OTHER = "other"
    KEY_STATICS = "statics"
    KEY_DYNAMICS = "dynamics"
    KEY_TRANSIENTS = "transients"

    BLOCK_TYPE_NONE = "none"
    BLOCK_TYPE_NORMAL = "normal"
    BLOCK_TYPE_MITOSIS = "mitosis"
    BLOCK_TYPE_RESOURCE = "resource"
    BLOCK_TYPE_IMPASSAblE = "impassable"

    GAME_OBJECT_KEY_ID = "id"
    GAME_OBJECT_KEY_OBJECT_ID = "objectId"
    GAME_OBJECT_KEY_TURN = "turn"
    GAME_OBJECT_KEY_TYPE = "type"
    GAME_OBJECT_KEY_OTHER = "other"
    GAME_OBJECT_KEY_TEAM_ID = "teamId"
    GAME_OBJECT_KEY_DURATION = "duration"
    GAME_OBJECT_KEY_POSITION = "position"

    INFO_KEY_TURN = "turn"
    INFO_KEY_TEAMS = "teams"
    INFO_KEY_VIEWS = "views"
    INFO_KEY_YOUR_INFO = "yourInfo"
    INFO_KEY_MAPSIZE = "mapSize"

    BLOCK_KEY_HEIGHT = "height"
    BLOCK_KEY_RESOURCE = "resource"

    CELL_KEY_ENERGY = "energy"

    GAME_OBJECT_TYPE_CELL = "cell"

    MESSAGE_TYPE_EVENT = "event"

    CELL_MIN_ENERGY_FOR_MITOSIS = 80
    CELL_MAX_ENERGY = 100
    CELL_GAIN_RATE = 15
    CELL_DEPTH_OF_FIELD = 2


class Event():
    TYPE_MITOSIS = "mitosis"
    TYPE_MOVE = "move"
    TYPE_GAIN_RESOURCE = "gainResource"

    def __init__(self, type1, id1, team_id=None):
        self.type = type1
        self.id = id1
        self.args = []
        self.team_id = team_id

    def add_arg(self, arg):
        self.args.append(arg)

    def to_message(self):
        return {
            Constants.GAME_OBJECT_KEY_TYPE: self.type,
            Constants.GAME_OBJECT_KEY_OBJECT_ID: self.id,
            Constants.KEY_ARGS: [arg for arg in self.args]
        }
