import socket
import json


class Network():
    def __init__(self, ip, port, token, message_handler):
        self.receive_flag = True
        self.ip = ip
        self.port = port
        self.token = token
        self.message_handler = message_handler
        self.result = b''
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.is_connected = False

    def connect(self):
        try:
            self.s.connect((self.ip, self.port))
            self.send({"name": "token", "args": [self.token]})
            init = self.receive()
            if init["name"] == "wrong token":
                raise ConnectionRefusedError("wrong token")
            elif not init["name"] == "init":
                self.close()
                raise IOError("first message was not init")
        except Exception as e:
            print("error while connecting to server", e)
            return
        print("connected to server!")
        self.is_connected = True
        self.message_handler(init)
        self.start_receiving()

    def send(self, message):
        self.s.send(json.dumps(message).encode('UTF-8'))
        self.s.send(b'\x00')

    def receive(self):
        while self.receive_flag:
            self.result += self.s.recv(1024)
            if b'\x00' in self.result:
                ans = json.loads(self.result[:self.result.index(b'\x00')].decode('UTF-8'))
                self.result = self.result[self.result.index(b'\x00')+1:]
                return ans

    def start_receiving(self):
        import threading

        def run():
            while self.receive_flag:
                self.message_handler(self.receive())

        tr = threading.Thread(target=run(), daemon=True)
        tr.start()

    def terminate(self):
        self.receive_flag = False

    def close(self):
        self.terminate()
        self.s.close()